# -*- coding: utf-8 -*-
""" xbstream library """

import errno
import logging
import os
import struct
import sys

LOG = logging.getLogger('xbstream')

XB_STREAM_CHUNK_MAGIC     = 'XBSTCK01'
XB_STREAM_CHUNK_MAGIC_LEN = len(XB_STREAM_CHUNK_MAGIC)

# Magic + flags + type + path len
CHUNK_HEADER_CONSTANT_LEN = XB_STREAM_CHUNK_MAGIC_LEN + 1 + 1 + 4
PATH_LENGTH_OFFSET        = XB_STREAM_CHUNK_MAGIC_LEN + 1 + 1
PATH_LENGTH_OFFSET_END    = PATH_LENGTH_OFFSET + 4
XB_CHUNK_TYPE_EOF         = 'E'
CHUNK_TYPE_OFFSET         = XB_STREAM_CHUNK_MAGIC_LEN + 1


class XBStreamObject(object):
    def __init__(self, init = True):
        self.buffer         = ""
        self.buffer_size    = 0
        self.filled_size    = 0
        self.tx_size        = 0
        self.chunk_tx       = False
        self.name           = None
        self.name_len       = 0
        self.magic_verified = False
        self.chunk_path_len = 0
        self.chunk_type     = None
        self.payload_size   = 0
        self.chunk_size     = CHUNK_HEADER_CONSTANT_LEN
        self.tx_started     = False


class XBStreamRead(object):
    EOF              = False
    FILE_CHUNK_COUNT = {}

    def __init__(self, rx_func = sys.stdin.read, tx_func = sys.stdout.write, xb_obj = None):
        self.rx_func    = rx_func
        self.tx_func    = tx_func
        self.xb_obj     = xb_obj or XBStreamObject()
        self.tx_func    = open(os.devnull, 'w').write

    def __call__(self, xb_obj = None):
        if xb_obj:
            self.xb_obj = xb_obj

        if self.xb_obj.chunk_tx:
            self.xb_obj.__init__()

        if self.xb_obj.filled_size < self.xb_obj.chunk_size:
            x = self.rx_func(self.xb_obj.chunk_size - self.xb_obj.filled_size)
            if x:
                self.xb_obj.buffer      += x
                self.xb_obj.filled_size += len(x)
                self.buffer_updated()
            else:
                self.EOF = True

    def is_streaming(self):
        return not self.EOF

    def tx_start(self):
        self.xb_obj.tx_started = True

        if self.xb_obj.filled_size == self.xb_obj.tx_size \
           and self.xb_obj.tx_size < self.xb_obj.chunk_size \
           and not self.EOF:
            x = None
            while True:
                try:
                    x = self.rx_func(self.xb_obj.chunk_size - self.xb_obj.filled_size)
                    break
                except errno.EAGAIN:
                    continue

            if x:
                self.xb_obj.buffer      += x
                self.xb_obj.filled_size += len(x)
                self.buffer_updated()
            else:
                self.EOF = True

        real_size = self.xb_obj.filled_size - self.xb_obj.tx_size

        tx_size = self.tx_func(self.xb_obj.buffer)
        if tx_size is not None:
            real_size = min(tx_size, self.xb_obj.filled_size - self.xb_obj.tx_size)

        self.xb_obj.tx_size += real_size

        if self.xb_obj.tx_size == self.xb_obj.chunk_size:
            self.tx_func(self.xb_obj.buffer)
            self.xb_obj.chunk_tx = True

    def buffer_updated(self):
        ready_for_tx = False

        if not self.xb_obj.magic_verified and self.xb_obj.filled_size >= CHUNK_HEADER_CONSTANT_LEN:
            if self.xb_obj.buffer[0:XB_STREAM_CHUNK_MAGIC_LEN] != XB_STREAM_CHUNK_MAGIC:
                raise ValueError("Error: magic excepted")
            self.xb_obj.magic_verified = True
            self.xb_obj.chunk_path_len = struct.unpack('I', self.xb_obj.buffer[PATH_LENGTH_OFFSET:PATH_LENGTH_OFFSET_END])[0]
            self.xb_obj.chunk_type     = self.xb_obj.buffer[CHUNK_TYPE_OFFSET]
            self.xb_obj.chunk_size     = CHUNK_HEADER_CONSTANT_LEN + self.xb_obj.chunk_path_len
            LOG.debug("buffer updated step 1: "
                      "chunk_path_len: %r, "
                      "chunk_type: %r, "
                      "chunk_size: %r, "
                      "filled_size: %r",
                      self.xb_obj.chunk_path_len,
                      self.xb_obj.chunk_type,
                      self.xb_obj.chunk_size,
                      self.xb_obj.filled_size)

            if self.xb_obj.chunk_type != XB_CHUNK_TYPE_EOF:
                self.xb_obj.chunk_size += 16

        if self.xb_obj.magic_verified \
           and self.xb_obj.payload_size == 0 \
           and self.xb_obj.chunk_type != XB_CHUNK_TYPE_EOF \
           and self.xb_obj.filled_size >= (CHUNK_HEADER_CONSTANT_LEN + self.xb_obj.chunk_path_len + 16):
            payload_len_offset       = CHUNK_HEADER_CONSTANT_LEN + self.xb_obj.chunk_path_len
            payload_len_offset_end   = CHUNK_HEADER_CONSTANT_LEN + self.xb_obj.chunk_path_len + 8
            self.xb_obj.payload_size = struct.unpack('L', self.xb_obj.buffer[payload_len_offset:payload_len_offset_end])[0]
            self.xb_obj.chunk_size   = self.xb_obj.payload_size + 4 + 16 + self.xb_obj.chunk_path_len + CHUNK_HEADER_CONSTANT_LEN
            self.xb_obj.name         = self.xb_obj.buffer[CHUNK_HEADER_CONSTANT_LEN:CHUNK_HEADER_CONSTANT_LEN + self.xb_obj.chunk_path_len]
            LOG.debug("buffer updated step 2: "
                      "name: %r, "
                      "chunk_size: %r, "
                      "payload_size: %r, "
                      "filled_size: %r",
                      self.xb_obj.name,
                      self.xb_obj.chunk_size,
                      self.xb_obj.payload_size,
                      self.xb_obj.filled_size)

        if self.xb_obj.magic_verified \
           and self.xb_obj.chunk_type == XB_CHUNK_TYPE_EOF \
           and self.xb_obj.filled_size >= (CHUNK_HEADER_CONSTANT_LEN + self.xb_obj.chunk_path_len):
            self.xb_obj.name = self.xb_obj.buffer[CHUNK_HEADER_CONSTANT_LEN:CHUNK_HEADER_CONSTANT_LEN + self.xb_obj.chunk_path_len]
            LOG.debug("buffer updated step 3: "
                      "name: %r, "
                      "chunk_size: %r, "
                      "filled_size: %r",
                      self.xb_obj.name,
                      self.xb_obj.chunk_size,
                      self.xb_obj.filled_size)

        if self.xb_obj.filled_size > 0 and self.xb_obj.filled_size == self.xb_obj.chunk_size:
            ready_for_tx = True
            LOG.debug("buffer updated step 4: "
                      "name: %r, "
                      "chunk_size: %r, "
                      "payload_size: %r, "
                      "filled_size: %r",
                      self.xb_obj.name,
                      self.xb_obj.chunk_size,
                      self.xb_obj.payload_size,
                      self.xb_obj.filled_size)

        if not self.xb_obj.tx_started and ready_for_tx:
            LOG.info("ready_for_tx: name: %r", self.xb_obj.name)
            self.tx_start()


if __name__ == '__main__':
    logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.DEBUG)
    xbread = XBStreamRead()
    try:
        while xbread.is_streaming():
            xbread()
    except KeyboardInterrupt:
        sys.stdout.flush()
    sys.exit(0)
